import copy
import json
from urllib.parse import urlparse

import yaml

from .env_desc import EnvDesc
from ..azure_utility import get_blob_container_client, get_blob_client

ENV_CONTAINER_NAME = "terraformoutput"
VERSION_CONTAINER_NAME = "versions"
SERVICE_VERSIONS_FOLDER_NAME = "service_versions"
DESC_TEMPLATES_FOLDER_NAME = "desc_templates"

BT_VERSION_KEY = "bt_version"


async def list_envs():
    blob_names = await _load_env_file_blob_names()
    return _extract_envs_from_blob_names(blob_names)


async def load_env_version(env_name: str):
    env_blob_data = await _load_env_blob_data(env_name)

    if env_blob_data is None:
        raise Exception(f"Environment not found ({env_name})")

    return _extract_version(env_blob_data)


async def load_env_desc(env_name: str):
    env_blob_data = await _load_env_blob_data(env_name)

    if env_blob_data is None:
        raise Exception(f"Environment not found ({env_name})")

    version = _extract_version(env_blob_data)
    desc_template = _load_desc_template(version)
    env_desc = _evaluate_desc_template(desc_template, env_blob_data)

    return env_desc


async def get_service_versions(version: str):
    version_blob = get_blob_client(
        VERSION_CONTAINER_NAME, f"{SERVICE_VERSIONS_FOLDER_NAME}/{version}.yaml"
    )

    if not version_blob.exists():
        raise Exception(f"Version not found ({version})")

    return yaml.safe_load(version_blob.download_blob().readall())


async def _load_env_file_blob_names():
    with get_blob_container_client(ENV_CONTAINER_NAME) as container_client:
        blob_names = []

        for blob in container_client.list_blobs():
            blob_names.append(blob.name)

        return blob_names


def _extract_envs_from_blob_names(blob_names):
    # Parse blob names and extract environments/stamps
    env_structure = {}

    for name in blob_names:
        name_parts = name.split("/")
        app_name = name_parts[0]
        env_name = name_parts[1]
        stamp_name = name_parts[2]

        if app_name not in env_structure:
            env_structure[app_name] = {}

        if env_name not in env_structure[app_name]:
            env_structure[app_name][env_name] = []

        if stamp_name not in env_structure[app_name][env_name]:
            env_structure[app_name][env_name].append(stamp_name)

    # Generate envs list
    envs = []

    for app_name in env_structure:
        for env_name in env_structure[app_name]:
            for stamp_name in env_structure[app_name][env_name]:
                envs.append(f"{app_name}.{env_name}.{stamp_name}")

    return envs


async def _load_env_blob_data(env_name: str):
    formatted_env_id = f"{env_name.replace('.', '/')}/@latest"

    with get_blob_container_client(ENV_CONTAINER_NAME) as container_client:
        tf_data = {}
        blob_names = []

        for blob in container_client.list_blobs(name_starts_with=formatted_env_id):
            blob_names.append(blob.name)

        if len(blob_names) == 0:
            return None

        for blob_name in blob_names:
            blob_client = container_client.get_blob_client(blob_name)
            blob_data = blob_client.download_blob()
            blob_obj = json.loads(blob_data.readall())
            tf_data.update(blob_obj)
            blob_client.close()

        return tf_data


def _load_desc_template(version: str):
    with get_blob_container_client(VERSION_CONTAINER_NAME) as container_client:
        blob_client = container_client.get_blob_client(
            f"{DESC_TEMPLATES_FOLDER_NAME}/{version}.yaml"
        )

        if not blob_client.exists():
            raise Exception(f"Desc template not found ({version})")

        service_version_yaml = yaml.safe_load(blob_client.download_blob().readall())
        return yaml.safe_dump(service_version_yaml)


def _evaluate_desc_template(template: str, env_blob_data: dict):
    # Generate extra data
    env_data = copy.deepcopy(env_blob_data)
    evaluated_template = template

    # env_data["tenant_id"] = "needed?"
    env_data["sub_id"] = _extract_sub_id(env_data)
    env_data["ingress_eh_url"] = _construct_ingress_eh_url(env_data)
    env_data["stamp_tag"] = _extract_stamp_tag(env_data)
    env_data["env_tag"] = _extract_env_tag(env_data)
    env_data["frontend_storage_name"] = _extract_frontend_storage_name(env_data)

    # Substitute variables in template
    var_start = "${"
    var_end = "}"

    var_start_index = evaluated_template.find(var_start)

    while var_start_index != -1:
        var_end_index = evaluated_template.find(var_end, var_start_index)
        var_name = evaluated_template[var_start_index + len(var_start) : var_end_index]
        var_value = env_data.get(var_name)

        if not var_value:
            raise Exception(f"Variable '{var_name}' not found in env data")

        # TODO: Consider doing it another way to avoid manually parsing while still using redundant `replace()`
        evaluated_template = evaluated_template.replace(
            f"{var_start}{var_name}{var_end}", var_value
        )
        var_start_index = evaluated_template.find(var_start, var_start_index)

    # Parse evaluated template
    return EnvDesc(yaml.safe_load(evaluated_template))


def _extract_sub_id(env_data: dict):
    if "common_plan_id" not in env_data:
        raise Exception("'common_plan_id' not found in environment data")

    common_plan_id = env_data["common_plan_id"]
    return common_plan_id.split("/")[2]


def _construct_ingress_eh_url(env_data: dict):
    if (
        "common_eventhub_namespace" not in env_data
        and "ingress_eventhub_name" not in env_data
    ):
        raise Exception(
            "'common_eventhub_namespace' or 'ingress_eventhub_name' not found in environment data"
        )

    eh_namespace = env_data["common_eventhub_namespace"]
    ingress_eh_name = env_data["ingress_eventhub_name"]

    return f"https://{eh_namespace}.servicebus.windows.net/{ingress_eh_name}"


def _extract_stamp_tag(env_data: dict):
    if (
        "environment" not in env_data
        and "shloc" not in env_data
        and "stamp" not in env_data
    ):
        raise Exception(
            "'environment' or 'shloc' or 'stamp' not found in environment data"
        )

    return f"{env_data['environment']}-{env_data['shloc']}-{env_data['stamp']}"


def _extract_env_tag(env_data: dict):
    if (
        "environment" not in env_data
        and "shloc" not in env_data
        and "stamp" not in env_data
        and "suffix" not in env_data
    ):
        raise Exception(
            "'environment' or 'shloc' or 'stamp' or 'suffix' not found in environment data"
        )

    return f"{env_data['environment']}-{env_data['shloc']}-{env_data['stamp']}{env_data['suffix']}"


def _extract_frontend_storage_name(env_data: dict):
    if "frontend_url" not in env_data:
        raise Exception("'frontend_url' not found in environment data")

    url = env_data["frontend_url"]

    return urlparse(url).netloc.split(".")[0]


def _extract_version(env_blob_data: dict):
    if BT_VERSION_KEY not in env_blob_data:
        raise Exception("'bt_version' not found in environment data")

    return env_blob_data[BT_VERSION_KEY]



import json
from typing import Annotated

from cyclopts import App
from typer import Argument

from btpy.core.migrations.migration_loader import (
    list_migrations,
    get_migration,
    find_migrations,
)
from btpy.print_utility import DateTimeJsonEncoder

app = App()


@app.command(name="list")
async def list_migrations_cmd():
    migrations = await list_migrations()

    for migration in migrations:
        print(migration)


@app.command(name="show")
async def show_migration_cmd(migration_id: str):
    try:
        migration = await get_migration(migration_id)

        migration_json = json.dumps(migration, indent=4, cls=DateTimeJsonEncoder)
        print(migration_json)
    except Exception as e:
        print(e)


@app.command(name="find")
async def find_migrations_cmd(
    from_version: Annotated[str, Argument(help="wtf??")],
    to_version: Annotated[str, Argument()],
):
    try:
        matched_migrations = await find_migrations(from_version, to_version)

        for migration in matched_migrations:
            print(json.dumps(migration, indent=4, cls=DateTimeJsonEncoder))
    except Exception as e:
        print(e)



import os
import sys

import yaml
from pathlib import Path

ENV_DATA_STORAGE_NAME_ENV_VAR_KEY = "BTPY_ENV_DATA_STORAGE_NAME"
ENV_DATA_STORAGE_NAME_CONFIG_KEY = "EnvDataStorageName"


def get_config_path(app_name: str) -> Path:
    if os.name == "nt":  # Windows
        return (
            Path(os.getenv("APPDATA", Path.home() / "AppData" / "Roaming")) / app_name
        )
    elif os.name == "posix":  # Unix-like systems (Linux, macOS, etc.)
        if sys.platform == "darwin":  # macOS
            return Path.home() / "Library" / "Application Support" / app_name
        else:  # Linux and other Unix-like systems
            return (
                Path(os.getenv("XDG_CONFIG_HOME", Path.home() / ".config")) / app_name
            )
    else:
        raise NotImplementedError(f"Unsupported platform: {os.name}")


def get_env_data_storage_name():
    # Try env var
    env_value = os.getenv(ENV_DATA_STORAGE_NAME_ENV_VAR_KEY)

    if env_value:
        return env_value

    # Try user config file
    conf = get_config_path("btpy")
    # conf = get_config_path("btpy").mkdir(parents=True, exist_ok=True)
    config_file_path = os.path.expanduser(os.path.join(conf, "config.yaml"))

    if not os.path.exists(config_file_path):
        raise FileNotFoundError(
            f"Failed to read env data storage name from file: {config_file_path} \
        (can also be set through env var: {ENV_DATA_STORAGE_NAME_ENV_VAR_KEY})"
        )

    with open(config_file_path, "r") as config_file:
        config_yaml = yaml.safe_load(config_file)
        config_value = config_yaml[ENV_DATA_STORAGE_NAME_CONFIG_KEY]

        if config_value:
            return config_value



import json
from dda_python_terraform import Terraform


def get_output():
    t = Terraform()
    return_code, out, err = t.output_cmd("-json", capture_output=True)
    return json.loads(out)


def clean_output(output):
    return {key: value["value"] for key, value in output.items()}



from cyclopts import App
import importlib.metadata

from btpy.commands import env_commands
from btpy.commands import software_commands
from btpy.commands import iac_commands


def get_version():
    return importlib.metadata.version("beacontower-btpy")


app = App(version=get_version)

app.command(env_commands.app, name="env")
app.command(software_commands.app, name="software")
app.command(iac_commands.app, name="iac")

if __name__ == "__main__":
    app()



import shlex
import subprocess
import json


def az(command: str):
    args = ["az", *shlex.split(command), "--output", "json"]
    program = subprocess.run(args, capture_output=True, text=True)

    return program.returncode, json.loads(program.stdout), program.stderr



from btpy.core.envs.env_desc import ResourceDesc, CloudResourceType
from btpy.core.resource_clients.azure.azure_appservice_client import (
    AzureAppServiceClient,
)
from btpy.core.resource_clients.azure.azure_function_client import AzureFunctionClient
from btpy.core.resource_clients.azure.azure_storage_web_client import (
    AzureStorageWebClient,
)
from btpy.core.resource_clients.azure.azure_vm_client import AzureVmClient
from btpy.core.resource_clients.resource_client import ResourceClient


def get_resource_client(desc: ResourceDesc) -> ResourceClient:
    if desc.cloud_resource.type == CloudResourceType.AzureAppService:
        return AzureAppServiceClient(desc)
    if desc.cloud_resource.type == CloudResourceType.AzureVm:
        return AzureVmClient(desc)
    elif desc.cloud_resource.type == CloudResourceType.AzureFunction:
        return AzureFunctionClient(desc)
    elif desc.cloud_resource.type == CloudResourceType.AzureStorageWeb:
        return AzureStorageWebClient(desc)
    else:
        raise ValueError("Invalid CloudResource type")



from cyclopts import App
from btpy.iac import misc
from rich.progress import Progress, TimeElapsedColumn
from azure.storage.blob import BlobServiceClient
from btpy.core.azure_utility import get_azure_credential
from btpy.az import az


app = App()

logger = misc.setup_logging()


def fetch_all_versions():
    exit_code, result_dict, logs = az(
        "acr repository list -n beacontower --output json"
    )
    if exit_code != 0:
        logger.error(logs)
        exit(1)
    logger.debug(result_dict)
    release_dict = {}

    with Progress(
        *Progress.get_default_columns(),
        TimeElapsedColumn(),
    ) as progress:
        task = progress.add_task(
            "fetching versions for repositories",
            filename="repositories",
            start=True,
            total=len(result_dict) + 1,
        )
        for repo in result_dict:
            exit_code, result_dict, logs = az(
                f"acr repository show-tags -n beacontower --repository {repo} --output json --top 5 --orderby time_desc"
            )
            if exit_code != 0:
                logger.error(logs)
                exit(1)
            logger.debug(result_dict)
            for tag in result_dict:
                logger.info(f"{repo}:{tag}")
            repovalue = f"{repo}:{result_dict[1]}"
            container_name = repo.replace("beacontower-", "")
            release_dict[f"{container_name}-version"] = repovalue
            progress.update(task, advance=1)

        exit_code, result_dict, logs = az(
            'storage blob list --container-name "beacontower-webclient" --connection-string "DefaultEndpointsProtocol=https;AccountName=storbtrelease;AccountKey=syy2XekWvISccZFWQ+ivfTHR0XY6EmTkAaErAv9H4kDPa6DTx4mIbsDooRDQQJbXZl6uFVlnlL5tfL8x6dI5uw==;EndpointSuffix=core.windows.net;" --output json --query "[].name"'
        )
        # remove the .zip extension on each item and store it in a new list
        versions = []
        for item in result_dict:
            versions.append(item.replace(".zip", ""))

        logger.debug(versions)
        # filter out versions that are not semantically versioned
        versions = [version for version in versions if version.count(".") == 2]
        # sort the list containing semantically versioned strings
        versions.sort(key=lambda x: tuple(map(int, x.split("."))))
        logger.debug(versions)
        release_dict["frontend-version"] = versions[-1]
        progress.update(task, advance=1)
    # # pretty print the dictionary
    # for key, value in release_dict.items():
    #     print(f"{key}: {value}")
    return release_dict


@app.command
def auto_tfvars():
    release_dict = fetch_all_versions()
    misc.write_json("beacontower-versions.auto.tfvars.json", release_dict)


# @app.command
# def compare(
#     env1: str = typer.Option(..., help="The first environment to compare"),
#     env2: str = typer.Option(..., help="The second environment to compare"),
# ):
#     logger.info(f"Comparing {env1} and {env2}")


@app.command
def release(iac: bool, settings: bool, code: bool, all: bool, upload: bool):
    """create a new release of beacontower software

    Parameters
    ----------
    iac : bool
        release the IAC
    settings : bool
        release the settings
    code : bool
        release the code
    all : bool
        release everything
    upload : bool
        upload the release to the cloud
    """
    if all:
        iac = True
        settings = True
        code = True
    if iac:
        logger.info("Releasing IAC")
    if settings:
        logger.info("Releasing settings")
    if code:
        logger.info("Releasing code")
    if upload:
        logger.info("Uploading release")


@app.command
def list():
    storage = "stgbttfdata"
    credentials = get_azure_credential()
    client = BlobServiceClient(
        f"https://{storage}.blob.core.windows.net",
        credential=credentials,
    )
    containers = client.list_containers()
    print("List all containers")
    print("-------------------")
    for container in containers:
        print(container["name"])


if __name__ == "__main__":
    app(["list"])



import json
import chevron
import os

from btpy import misc

filename = "customer.json"
appjs = "application.json"


def get_directory():
    c = misc.find_file_in_parent_directories(appjs)
    return misc.get_directory_of_file(c)


def get_data():
    c = misc.find_file_in_parent_directories(appjs)
    with open(c, "r") as f:
        return json.load(f)


def genenerate_tfbackend_from_template(data):
    """
    Generate tfbackend.tf from template
    data should contain
    - resource_group_name
    - storage_account_name
    - container_name
    - key
    """
    appdir = get_directory()
    with open(os.path.join(appdir, "templates", "tfbackend.mustache"), "r") as f:
        return chevron.render(f, data)


def genenerate_tfproviders_from_template(data):
    """
    Generate tfproviders.tf from template
    data should contain
     - providers - list of providers with object containing
        - name
        - version
        - source
    """
    appdir = get_directory()
    with open(os.path.join(appdir, "templates", "tfproviders.mustache"), "r") as f:
        return chevron.render(f, data)



import sys
from pathlib import Path


def to_absolute_path(relative_path: str) -> Path:
    if getattr(sys, "frozen", False) and hasattr(sys, "_MEIPASS"):
        # noinspection PyProtectedMember
        script_dir = Path(sys._MEIPASS)
    else:
        script_dir = Path(__file__).parent

    return script_dir / relative_path



import json
from datetime import datetime

from btpy.core.resource_clients.resource_client import ResourceStatus


def print_status(status: ResourceStatus):
    if status == ResourceStatus.Running:
        color = "[green]"
    elif status == ResourceStatus.Stopped:
        color = "[red]"
    elif (
        status == ResourceStatus.Restarting
        or status == ResourceStatus.Unknown
        or status == ResourceStatus.Mixed
    ):
        color = "[yellow]"
    else:
        raise ValueError(f"Unknown status: {status}")

    return f"{color}{status.name}"


class DateTimeJsonEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime):
            return o.isoformat()
        return super().default(o)



# Import the certificate classes
from sshkey_tools.keys import (
    RsaPrivateKey,
    EcdsaPrivateKey,
    Ed25519PrivateKey,
    EcdsaCurves,
    PrivateKey,
)
import typer

app = typer.Typer()


@app.command()
def keypair(
    key_type: str = typer.Option("rsa", help="The type of key to generate"),
    key_size: int = typer.Option(2048, help="The size of the key to generate"),
    output_file: str = typer.Option("id_rsa", help="The file to save the key to"),
    password: str = typer.Option("", help="The password to encrypt the key with"),
):
    """
    Generate a new keypair
    """
    if key_type == "rsa":
        rsa_priv = RsaPrivateKey.generate(key_size)
        rsa_priv.to_file(output_file, password)
    elif key_type == "ecdsa":
        ecdsa_priv = EcdsaPrivateKey.generate(EcdsaCurves.NISTP256)
        ecdsa_priv.to_file(output_file, password)
    elif key_type == "ed25519":
        ed25519_priv = Ed25519PrivateKey.generate()
        ed25519_priv.to_file(output_file, password)


@app.command()
def public_key(
    key_type: str = typer.Option("rsa", help="The type of key to generate"),
    input_file: str = typer.Option("id_rsa", help="The file to read the key from"),
    output_file: str = typer.Option("id_rsa.pub", help="The file to save the key to"),
):
    """
    Generate a public key from a private key
    """
    priv = PrivateKey.from_file(input_file)
    pub = priv.public_key()
    pub.to_file(output_file)


@app.command()
def download_keypair():
    """
    Download a keypair from a remote server
    """
    pass


if __name__ == "__main__":
    app()



from urllib.parse import urljoin

from azure.core.exceptions import ResourceNotFoundError
from azure.mgmt.web.aio import WebSiteManagementClient
from azure.mgmt.web.v2022_09_01.models import Site

from btpy.core.azure_utility import get_azure_credential
from btpy.core.envs.env_desc import ResourceDesc
from btpy.core.resource_clients.resource_client import ResourceClient, ResourceStatus


class AzureAppServiceClient(ResourceClient):
    def __init__(self, desc: ResourceDesc):
        credential = get_azure_credential()

        self.desc = desc
        self.client = WebSiteManagementClient(
            credential=credential, subscription_id=desc.cloud_resource.sub_id
        )

    async def close(self):
        await self.client.close()

    async def status(self):
        try:
            app_client: Site = await self.client.web_apps.get(
                self.desc.cloud_resource.group, self.desc.cloud_resource.name
            )

            return ResourceStatus.from_str(app_client.state)
        except ResourceNotFoundError:
            return ResourceStatus.Unknown

    async def get_api_url(self):
        app_client: Site = await self.client.web_apps.get(
            self.desc.cloud_resource.group, self.desc.cloud_resource.name
        )

        return urljoin(
            f"https://{app_client.default_host_name}/api", self.desc.api_url_addition
        )

    async def get_log_level(self) -> str:
        if self.desc.settings is None:
            return await self._get_log_level_from_appsettings()
        else:
            raise NotImplementedError("Custom settings not implemented yet")

    async def _get_log_level_from_appsettings(self) -> str:
        app_settings = await self.client.web_apps.list_application_settings(
            self.desc.cloud_resource.group, self.desc.cloud_resource.name
        )

        return app_settings.properties.get("LogLevel", "info")



import yaml

from btpy.core.azure_utility import get_blob_container_client, get_blob_client
from btpy.core.envs.env_desc_loader import get_service_versions

MIGRATIONS_CONTAINER_NAME = "migrations"

APPLIES_TO_KEY = "applies_to"
SERVICE_KEY = "service"
VERSION_KEY = "version"


async def list_migrations() -> list[str]:
    migrations_container_client = get_blob_container_client(MIGRATIONS_CONTAINER_NAME)
    migration_id_it = migrations_container_client.list_blob_names()
    migration_names = []

    for migration_id in migration_id_it:
        if "/" not in migration_id:
            migration_blob = get_blob_client(MIGRATIONS_CONTAINER_NAME, migration_id)
            migration = yaml.safe_load(migration_blob.download_blob().readall())

            migration_names.append(to_pretty_migration_name(migration))

    return migration_names


async def get_migration(migration_id: str) -> dict:
    migration_blob = get_blob_client(MIGRATIONS_CONTAINER_NAME, f"{migration_id}.yaml")

    if not migration_blob.exists():
        raise Exception(f"Migration does not exist ({migration_id})")

    return yaml.safe_load(migration_blob.download_blob().readall())


async def find_migrations(from_version: str, to_version: str):
    service_version_changes = _merge_service_versions(
        await get_service_versions(from_version), await get_service_versions(to_version)
    )
    migrations = []
    migration_ids = set()
    all_migrations = await _load_all_migrations()

    for service, versions in service_version_changes.items():
        # Unhandled cases:
        # - If a removed service requires cleanup of database, how do you apply this with `applies_to` list?
        # - If a previously existed service is reintroduced and requires a migration, `from_version` will probably
        # be `None`.
        service_migrations = await find_service_migrations(
            all_migrations, service, versions[0], versions[1]
        )

        for service_migration in service_migrations:
            if service_migration["id"] in migration_ids:
                continue

            migration_ids.add(service_migration["id"])
            migrations.append(service_migration)

    return migrations


async def find_service_migrations(
    all_migrations: list[dict], service: str, from_version: str, to_version: str
):
    if from_version is None or to_version is None:
        return []
    if from_version > to_version:
        raise Exception(
            f"from_version must not be higher than to_version ({from_version}, {to_version})"
        )
    if from_version == to_version:
        return []

    matched_migrations = []

    for migration in all_migrations:
        if _matches_service_version(migration, service, from_version, to_version):
            matched_migrations.append(migration)

    return matched_migrations


async def _load_all_migrations():
    migration_container = get_blob_container_client(MIGRATIONS_CONTAINER_NAME)
    migration_blob_name_it = migration_container.list_blob_names()
    migrations = []

    for migration_blob_name in migration_blob_name_it:
        if "/" not in migration_blob_name:
            migrations.append(
                await get_migration(_without_file_ext(migration_blob_name))
            )

    return migrations


def _matches_service_version(
    migration: dict, service: str, min_version: str, max_version: str
):
    if APPLIES_TO_KEY not in migration:
        raise Exception(f"Invalid migration (missing property {APPLIES_TO_KEY})")

    for service_version in migration[APPLIES_TO_KEY]:
        if service_version[SERVICE_KEY] == service:
            version = service_version[VERSION_KEY]
            if min_version < version <= max_version:
                return True


def _merge_service_versions(from_versions: dict, to_versions: dict):
    service_version_changes = {}

    for service_name, version in from_versions.items():
        if service_name in service_version_changes:
            raise Exception(f"Duplicate service from-version found ({service_name})")

        service_version_changes[service_name] = (version, None)

    for service_name, version in to_versions.items():
        if service_name not in service_version_changes:
            service_version_changes[service_name] = (None, version)
            continue

        if service_version_changes[service_name][1] is not None:
            raise Exception(f"Duplicate service to-version found ({service_name})")

        service_version_changes[service_name] = (
            service_version_changes[service_name][0],
            version,
        )

    return service_version_changes


def to_pretty_migration_name(migration: dict):
    return f"{migration['id']} - {migration['name']}"


def _without_file_ext(filename: str):
    return filename.rsplit(".", 1)[0]



import asyncio

from cyclopts import App
from rich import print
from rich.progress import Progress, SpinnerColumn, TextColumn

from btpy.core.envs import env_desc_loader
from btpy.core.envs.env_client import EnvClient
from btpy.core.envs.env_desc_loader import load_env_version
from btpy.core.envs.env_migrator import migrate_env
from btpy.print_utility import print_status

app = App()


@app.command(name="list")
async def list_envs_cmd():
    env_list = await env_desc_loader.list_envs()

    for env in env_list:
        print(env)


@app.command(name="version")
async def env_version_cmd(env_name: str):
    try:
        version = await load_env_version(env_name)
        print(version)
    except Exception as e:
        print(e)


@app.command(name="status")
async def env_status_cmd(env_name: str):
    try:
        env_desc = await env_desc_loader.load_env_desc(env_name)

        if env_desc is None:
            print("Env not found")
            return
    except Exception as e:
        print(e)
        return

    env_client = EnvClient(env_desc)
    resource_clients = env_client.get_resource_clients()

    with Progress(
        SpinnerColumn(),
        TextColumn("{task.description}"),
        TextColumn("{task.fields[result]}"),
    ) as table:
        loop = asyncio.get_event_loop()
        tasks = [
            (
                resource_client,
                loop.create_task(_get_formatted_status(resource_client.client)),
                table.add_task(f"[blue]{resource_client.name}[/blue]", result=""),
            )
            for resource_client in resource_clients
        ]

        while not table.finished:
            for resource_client, task, table_task in tasks:
                if task.done():
                    table.update(
                        table_task, total=100, completed=100, result=task.result()
                    )

            await asyncio.sleep(0.1)

    [await resource_client.client.close() for resource_client in resource_clients]


@app.command(name="upgrade")
async def upgrade_env_cmd(env_name: str, to_version: str):
    await migrate_env(env_name, to_version)


async def _get_formatted_status(resource_client):
    status = await resource_client.status()
    return print_status(status)



import json
from btpy import misc

envjs = "environment.json"


def get_directory():
    c = misc.find_file_in_parent_directories(envjs)
    return misc.get_directory_of_file(c)


def get_data():
    c = misc.find_file_in_parent_directories(envjs)
    with open(c, "r") as f:
        return json.load(f)



from enum import Enum


class EnvDesc:
    def __init__(self, desc_obj: dict):
        self.tenant = TenantDesc(desc_obj["tenant"])
        self.version = desc_obj["version"]
        self.default_user = DefaultUserDesc(desc_obj["defaultUser"])
        self.services = [ServiceDesc(service) for service in desc_obj["services"]]


class TenantDesc:
    def __init__(self, desc_obj: dict):
        self.id = desc_obj["id"]
        self.client_id = desc_obj["clientId"]
        self.sign_in_flow = desc_obj["signInFlow"]


class DefaultUserDesc:
    def __init__(self, desc_obj: dict):
        self.email = desc_obj["email"]
        self.password = desc_obj["password"]


class ServiceDesc:
    def __init__(self, desc_obj: dict):
        self.name = desc_obj["name"]
        self.resources = [ResourceDesc(resource) for resource in desc_obj["resources"]]
        # self.settings = load_service_settings_from_json(desc_obj.get("settings"))


class ResourceDesc:
    def __init__(self, desc_obj: dict):
        self.name = desc_obj["name"]
        self.cloud_resource = load_cloud_resource_desc_from_json(
            desc_obj["cloudResource"]
        )
        self.api_url_addition = desc_obj.get("apiUrlAddition")
        self.settings = desc_obj.get("settings")


class AzureAppServiceDesc:
    def __init__(self, desc_obj: dict):
        if desc_obj["type"] != CloudResourceType.AzureAppService.value:
            raise ValueError("Invalid CloudResource type")

        self.type = CloudResourceType.AzureAppService
        self.name = desc_obj["name"]
        self.group = desc_obj["group"]
        self.sub_id = desc_obj["subscriptionId"]
        # self.tenant_id = desc_obj["tenantId"]


class AzureFunctionDesc:
    def __init__(self, desc_obj: dict):
        if desc_obj["type"] != CloudResourceType.AzureFunction.value:
            raise ValueError("Invalid CloudResource type")

        self.type = CloudResourceType.AzureFunction
        self.name = desc_obj["name"]
        self.group = desc_obj["group"]
        self.sub_id = desc_obj["subscriptionId"]
        # self.tenant_id = desc_obj["tenantId"]


class AzureVmDesc:
    def __init__(self, desc_obj: dict):
        if desc_obj["type"] != CloudResourceType.AzureVm.value:
            raise ValueError("Invalid CloudResource type")

        self.type = CloudResourceType.AzureVm
        self.name = desc_obj["name"]
        self.group = desc_obj["group"]
        self.sub_id = desc_obj["subscriptionId"]
        # self.tenant_id = desc_obj["tenantId"]
        self.sub_desc = load_vm_sub_desc_from_json(desc_obj["subDesc"])


class AzureStorageWebDesc:
    def __init__(self, desc_obj: dict):
        if desc_obj["type"] != CloudResourceType.AzureStorageWeb.value:
            raise ValueError("Invalid CloudResource type")

        self.type = CloudResourceType.AzureStorageWeb
        self.name = desc_obj["name"]
        self.group = desc_obj["group"]
        self.sub_id = desc_obj["subscriptionId"]
        # self.tenant_id = desc_obj["tenantId"]


class CloudResourceType(Enum):
    AzureAppService = "AzureAppService"
    AzureFunction = "AzureFunction"
    AzureVm = "AzureVm"
    AzureStorageWeb = "AzureStorageWeb"


class VmDockerComposeSubDesc:
    def __init__(self, desc_obj: dict):
        if desc_obj["type"] != VmSubDescType.DockerCompose.value:
            raise ValueError("Invalid VmSubDesc type")

        self.type = VmSubDescType.DockerCompose
        self.path = desc_obj["path"]
        self.config_files = desc_obj.get("configFiles")


class VmSubDescType(Enum):
    DockerCompose = "DockerCompose"


def load_vm_sub_desc_from_json(desc_obj: dict):
    desc_type = desc_obj["type"]

    if desc_type == "DockerCompose":
        return VmDockerComposeSubDesc(desc_obj)


def load_cloud_resource_desc_from_json(desc_obj: dict):
    desc_type = desc_obj["type"]

    if desc_type == "AzureAppService":
        return AzureAppServiceDesc(desc_obj)
    elif desc_type == "AzureFunction":
        return AzureFunctionDesc(desc_obj)
    elif desc_type == "AzureVm":
        return AzureVmDesc(desc_obj)
    elif desc_type == "AzureStorageWeb":
        return AzureStorageWebDesc(desc_obj)
    else:
        raise ValueError(f"Invalid CloudResource type: {desc_type}")



import os
import subprocess
import tempfile
import venv

from rich import print

from btpy.core.azure_utility import get_blob_client, get_blob_container_client
from btpy.core.envs.env_desc import EnvDesc
from btpy.core.migrations.migration_loader import MIGRATIONS_CONTAINER_NAME


async def run_migration(env_desc: EnvDesc, migration: dict):
    # TODO: Verify that migration files exist

    with tempfile.TemporaryDirectory() as temp_dir:
        try:
            _create_venv(temp_dir)
            await _download_migration_files(migration, temp_dir)
            _install_requirements(temp_dir)
            _execute_migration_python_files(temp_dir)

            # TODO: Handle output and errors in a good way

            return True
        except Exception as e:
            print(e)
            return False


async def _download_migration_files(migration: dict, temp_dir: str):
    print("[blue]Downloading migration files...", end=" ")

    try:
        migration_file_blob_it = get_blob_container_client(
            MIGRATIONS_CONTAINER_NAME
        ).list_blobs(name_starts_with=f"{migration['id']}/")

        for migration_file_blob_entry in migration_file_blob_it:
            migration_file_blob = get_blob_client(
                MIGRATIONS_CONTAINER_NAME, migration_file_blob_entry.name
            )
            migration_file_content = migration_file_blob.download_blob().readall()
            migration_file_path = os.path.join(
                temp_dir, os.path.basename(migration_file_blob_entry.name)
            )

            with open(migration_file_path, "w") as migration_file:
                migration_file.write(migration_file_content.decode("utf-8"))

        print("[green]Done")
    except Exception as e:
        print(f"[red]Failed ({e})")
        return None


def _create_venv(temp_dir: str) -> str:
    print("[blue]Creating venv...", end=" ")

    try:
        venv_dir = os.path.join(temp_dir, ".venv")
        venv.create(venv_dir, with_pip=True)

        # with_pip failed every time earlier, so I went with this
        # python_bin = os.path.join(venv_dir, "Scripts", "python")
        # subprocess.check_call([python_bin, "-m", "ensurepip", "--upgrade", "--default-pip"])

        print("[green]Done")
        return venv_dir
    except Exception as e:
        print(f"[red]Failed ({e})")
        raise


def _install_requirements(temp_dir: str):
    requirements_file = os.path.join(temp_dir, "requirements.txt")

    if not os.path.exists(requirements_file):
        return

    print("[blue]Installing requirements...", end=" ")

    result = subprocess.run(
        ["pip", "install", "-r", requirements_file], capture_output=True, text=True
    )

    if result.returncode != 0:
        print(f"[red]Failed ({result.stderr})")
        return

    print("[green]Done")


def _execute_migration_python_files(temp_dir: str):
    print("[blue]Executing python file...", end=" ")

    entry_point_file = os.path.join(temp_dir, "main.py")
    result = subprocess.run(
        ["python", entry_point_file], capture_output=True, text=True
    )
    # result = subprocess.run(["python", entry_point_file])

    if result.returncode != 0:
        # print(f"[red]Failed ({result.stderr})")
        print("[red]Failed")
        return

    print("[green]Done")



from azure.identity import AzureCliCredential
from azure.storage.blob import BlobServiceClient, ContainerClient

from btpy.configuration.config import get_env_data_storage_name

_blob_service_client = None


def get_azure_credential():
    return AzureCliCredential()
    # return DefaultAzureCredential()


def to_azure_appsetting_format(setting_value: str) -> str:
    return setting_value.replace(":", "__")


def from_azure_appsetting_format(setting_value: str) -> str:
    return setting_value.replace("__", ":")


def get_blob_service_client() -> BlobServiceClient:
    global _blob_service_client

    if _blob_service_client is None:
        credential = get_azure_credential()
        blob_account_url = (
            f"https://{get_env_data_storage_name()}.blob.core.windows.net"
        )
        _blob_service_client = BlobServiceClient(
            blob_account_url, credential=credential
        )

    return _blob_service_client


def get_blob_container_client(container_name: str) -> ContainerClient:
    blob_service_client = get_blob_service_client()
    return blob_service_client.get_container_client(container_name)


def get_blob_client(container_name: str, blob_name: str):
    container_client = get_blob_container_client(container_name)
    return container_client.get_blob_client(blob_name)


def to_blob_index_format(tag: str) -> str:
    return tag.replace("-", "_")
